#ifndef CAPTURE_GNSS_SINGLE_DIFF_H_
#define CAPTURE_GNSS_SINGLE_DIFF_H_

//Wolf includes
#include "gnss/feature/feature_gnss_single_diff.h"
#include "core/capture/capture_base.h"

//std includes
//

namespace wolf {
    
WOLF_PTR_TYPEDEFS(CaptureGnssSingleDiff);

//class CaptureGnssSingleDiff
class CaptureGnssSingleDiff : public CaptureBase
{
    protected:
        Eigen::Vector3s data_; ///< Raw data.
        Eigen::Matrix3s data_covariance_; ///< Noise of the capture.
        FrameBasePtr    origin_frame_ptr_;

    public:
        CaptureGnssSingleDiff(const TimeStamp& _ts,
                              SensorBasePtr _sensor_ptr,
                              const Eigen::Vector3s& _data,
                              const Eigen::Matrix3s& _data_covariance,
                              const FrameBasePtr& _origin_frame_ptr) :
            CaptureBase("GNSS SINGLE DIFFERENCE", _ts, _sensor_ptr),
            data_(_data),
            data_covariance_(_data_covariance),
            origin_frame_ptr_(_origin_frame_ptr)
        {};
        virtual ~CaptureGnssSingleDiff(){};
        const Eigen::Vector3s& getData() const;
        const Eigen::Matrix3s& getDataCovariance() const;
        void getDataAndCovariance(Eigen::Vector3s& data, Eigen::Matrix3s& data_cov) const;
        FrameBasePtr getOriginFrame() const;
};

inline const Eigen::Vector3s& CaptureGnssSingleDiff::getData() const
{
    return data_;
}

inline const Eigen::Matrix3s& CaptureGnssSingleDiff::getDataCovariance() const
{
    return data_covariance_;
}

inline void CaptureGnssSingleDiff::getDataAndCovariance(Eigen::Vector3s& data, Eigen::Matrix3s& data_cov) const
{
    data = data_;
    data_cov = data_covariance_;
}

inline FrameBasePtr CaptureGnssSingleDiff::getOriginFrame() const
{
    return origin_frame_ptr_;
}

} //namespace wolf
#endif
